# -*- coding: utf-8 -*-
"""Supports Dst values. Downloads data from NGDC.

Properties
----------
platform
    'sw'
name
    'dst'
tag
    None supported

Note
----
Will only work until 2057.

Download method should be invoked on a yearly frequency,
dst.download(date1, date2, freq='AS')

This material is based upon work supported by the
National Science Foundation under Grant Number 1259508.

Any opinions, findings, and conclusions or recommendations expressed in this
material are those of the author(s) and do not necessarily reflect the views
of the National Science Foundation.

"""

import datetime as dt
import ftplib
import numpy as np
import os
import pandas as pds

import pysat
from pysatSpaceWeather.instruments.methods import dst as mm_dst

logger = pysat.logger

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'dst'
tags = {'': ''}
inst_ids = {'': ['']}

# ----------------------------------------------------------------------------
# Instrument test attributes

_test_dates = {'': {'': dt.datetime(2007, 1, 1)}}

# Other tags assumed to be True
_test_download_travis = {'': {'': False}}

# ----------------------------------------------------------------------------
# Instrument methods


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.


    """

    self.acknowledgements = mm_dst.acknowledgements(self.name, self.tag)
    self.references = mm_dst.references(self.name, self.tag)
    logger.info(self.acknowledgements)
    return


def clean(self):
    """ Cleaning function for Dst

    Note
    ----
    No necessary for the Dst index

    """
    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag=None, inst_id=None):
    """Load Kp index files

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames
    tag : str or NoneType
        tag or None (default=None)
    inst_id : str or NoneType
        satellite id or None (default=None)

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data
    pysat.Meta
        Object containing metadata such as column names and units

    Note
    ----
    Called by pysat. Not intended for direct use by user.


    """

    data = pds.DataFrame()

    for filename in fnames:
        # need to remove date appended to dst filename
        fname = filename[0:-11]
        all_data = []
        with open(fname) as open_f:
            lines = open_f.readlines()
            idx = 0
            # check if all lines are good
            max_lines = 0
            for line in lines:
                if len(line) > 1:
                    max_lines += 1
            yr = np.zeros(max_lines * 24, dtype=int)
            mo = np.zeros(max_lines * 24, dtype=int)
            day = np.zeros(max_lines * 24, dtype=int)
            ut = np.zeros(max_lines * 24, dtype=int)
            dst = np.zeros(max_lines * 24, dtype=int)
            for line in lines:
                if len(line) > 1:
                    temp_year = int(line[14:16] + line[3:5])
                    if temp_year > 57:
                        temp_year += 1900
                    else:
                        temp_year += 2000

                    yr[idx:idx + 24] = temp_year
                    mo[idx:idx + 24] = int(line[5:7])
                    day[idx:idx + 24] = int(line[8:10])
                    ut[idx:idx + 24] = np.arange(24)
                    temp = line.strip()[20:-4]
                    temp2 = [temp[4 * i:4 * (i + 1)] for i in np.arange(24)]
                    dst[idx:idx + 24] = temp2
                    idx += 24

            start = dt.datetime(yr[0], mo[0], day[0], ut[0])
            stop = dt.datetime(yr[-1], mo[-1], day[-1], ut[-1])
            dates = pds.date_range(start, stop, freq='H')

            new_data = pds.DataFrame(dst, index=dates, columns=['dst'])
            # pull out specific day
            new_date = dt.datetime.strptime(filename[-10:], '%Y-%m-%d')
            idx, = np.where((new_data.index >= new_date)
                            & (new_data.index < new_date
                               + pds.DateOffset(days=1)))
            new_data = new_data.iloc[idx, :]
            # add specific day to all data loaded for filenames
            all_data.append(new_data)
        # combine data together
        data = pds.concat(all_data, sort=True, axis=0)

    return data, pysat.Meta()


def list_files(tag=None, inst_id=None, data_path=None, format_str=None):
    """Return a Pandas Series of every file for chosen satellite data

    Parameters
    ----------
    tag : string or NoneType
        Denotes type of file to load.  Accepted types are '1min' and '5min'.
        (default=None)
    inst_id : string or NoneType
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    data_path : string or NoneType
        Path to data directory.  If None is specified, the value previously
        set in Instrument.files.data_path is used.  (default=None)
    format_str : string or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    pysat.Files.from_os : pysat._files.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    if data_path is not None:
        if tag == '':
            # files are by year, going to add date to yearly filename for
            # each day of the month. The load routine will load a month of
            # data and use the appended date to select out appropriate data.
            if format_str is None:
                format_str = 'dst{year:4d}.txt'
            out = pysat.Files.from_os(data_path=data_path,
                                      format_str=format_str)
            if not out.empty:
                out.loc[out.index[-1] + pds.DateOffset(years=1)
                        - pds.DateOffset(days=1)] = out.iloc[-1]
                out = out.asfreq('D', 'pad')
                out = out + '_' + out.index.strftime('%Y-%m-%d')
            return out
        else:
            raise ValueError(''.join(('Unrecognized tag name for Space ',
                                      'Weather Dst Index')))
    else:
        raise ValueError(''.join(('A data_path must be passed to the loading ',
                                  'routine for Dst')))


def download(date_array, tag, inst_id, data_path, user=None, password=None):
    """Routine to download Dst index data

    Parameters
    ----------
    tag : string or NoneType
        Denotes type of file to load.
        (default=None)
    inst_id : string or NoneType
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    data_path : string or NoneType
        Path to data directory.  If None is specified, the value previously
        set in Instrument.files.data_path is used.  (default=None)

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """
    # Connect to host, default port
    ftp = ftplib.FTP('ftp.ngdc.noaa.gov')

    # User anonymous, passwd anonymous@
    ftp.login()
    ftp.cwd('/STP/GEOMAGNETIC_DATA/INDICES/DST')

    # Data stored by year. Only download for unique set of input years.
    years = np.array([date.year for date in date_array])
    years = np.unique(years)
    for year in years:
        fname_root = 'dst{year:04d}.txt'
        fname = fname_root.format(year=year)
        saved_fname = os.path.join(data_path, fname)
        try:
            logger.info('Downloading file for {year:04d}'.format(year=year))
            with open(saved_fname, 'wb') as fp:
                ftp.retrbinary('RETR ' + fname, fp.write)
        except ftplib.error_perm as exception:
            if str(exception.args[0]).split(" ", 1)[0] != '550':
                raise
            else:
                # file not present
                os.remove(saved_fname)
                logger.info('File not available for {:04d}'.format(year))

    ftp.close()
    return



# -*- coding: utf-8 -*-.
"""Provides default routines for the real-time ACE instruments

"""


def acknowledgements(src):
    """Returns acknowledgements for the specified ACE instrument

    Parameters
    ----------
    src : string
        Name of the ACE instrument data source

    Returns
    -------
    ref : string
        Acknowledgements for the ACE instrument

    """

    ackn = {'sw': ''.join(['NOAA provided funds for the modification of the ',
                           'ACE transmitter to enable the broadcast of the ',
                           'real-time data and also funds to the instrument ',
                           'teams to provide the algorithms for processing ',
                           'the real-time data.'])}

    return ackn[src]


def references(iname):
    """Returns references for the specified ACE instrument

    Parameters
    ----------
    iname : string
        Name of the ACE instrument

    Returns
    -------
    ref : string
        Reference for the ACE instrument paper

    """

    refs = {'mag': "".join(["'The ACE Magnetic Field Experiment', ",
                            "C. W. Smith, M. H. Acuna, L. F. Burlaga, ",
                            "J. L'Heureux, N. F. Ness and J. Scheifele, ",
                            "Space Science Reviews, 86, 613-632 (1998)."]),
            'epam': ''.join(['Gold, R. E., S. M. Krimigis, S. E. Hawkins, ',
                             'D. K. Haggerty, D. A. Lohr, E. Fiore, ',
                             'T. P. Armstrong, G. Holland, L. J. Lanzerotti,',
                             ' Electron, Proton and Alpha Monitor on the ',
                             'Advanced Composition Explorer Spacecraft, ',
                             'Space Sci. Rev., 86, 541, 1998.']),
            'swepam': ''.join(['McComas, D., Bame, S., Barker, P. et al. ',
                               'Solar Wind Electron Proton Alpha Monitor ',
                               '(SWEPAM) for the Advanced Composition ',
                               'Explorer. Space Science Reviews 86, 563–612',
                               ' (1998). ',
                               'https://doi.org/10.1023/A:1005040232597']),
            'sis': ''.join(['Stone, E., Cohen, C., Cook, W. et al. The ',
                            'Solar Isotope Spectrometer for the Advanced ',
                            'Composition Explorer. Space Science Reviews ',
                            '86, 357–408 (1998). ',
                            'https://doi.org/10.1023/A:1005027929871'])}

    return refs[iname]



# -*- coding: utf-8 -*-
"""Supports F10.7 index values. Downloads data from LASP and the SWPC.

Properties
----------
platform
    'sw'
name
    'f107'
tag
    - '' LASP F10.7 data (downloads by month, loads by day)
    - 'all' All LASP standard F10.7
    - 'prelim' Preliminary SWPC daily solar indices
    - 'daily' Daily SWPC solar indices (contains last 30 days)
    - 'forecast' Grab forecast data from SWPC (next 3 days)
    - '45day' 45-Day Forecast data from the Air Force

Note
----
The forecast data is stored by generation date, where each file contains the
forecast for the next three days. Forecast data downloads are only supported
for the current day. When loading forecast data, the date specified with the
load command is the date the forecast was generated. The data loaded will span
three days. To always ensure you are loading the most recent data, load
the data with tomorrow's date.
::

    f107 = pysat.Instrument('sw', 'f107', tag='forecast')
    f107.download()
    f107.load(date=f107.tomorrow())



The forecast or prelim data should not be used with the data padding option
available from pysat.Instrument objects. The 'all' tag shouldn't be used either,
no other data available to pad with.

Warnings
--------
The 'forecast' F10.7 data loads three days at a time. The data padding feature
and multi_file_day feature available from the pyast.Instrument object
is not appropriate for 'forecast' data.

"""

import datetime as dt
import ftplib
import json
import numpy as np
import os
import requests
import sys
import warnings

import pandas as pds
import pysat

from pysatSpaceWeather.instruments.methods import f107 as mm_f107

logger = pysat.logger

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'f107'
tags = {'': 'Daily LASP value of F10.7',
        'all': 'All LASP F10.7 values',
        'prelim': 'Preliminary SWPC daily solar indices',
        'daily': 'Daily SWPC solar indices (contains last 30 days)',
        'forecast': 'SWPC Forecast F107 data next (3 days)',
        '45day': 'Air Force 45-day Forecast'}

# dict keyed by inst_id that lists supported tags for each inst_id
inst_ids = {'': ['', 'all', 'prelim', 'daily', 'forecast', '45day']}

# dict keyed by inst_id that lists supported tags and a good day of test data
# generate todays date to support loading forecast data
now = dt.datetime.now()
today = dt.datetime(now.year, now.month, now.day)
tomorrow = today + pds.DateOffset(days=1)

# ----------------------------------------------------------------------------
# Instrument test attributes

_test_dates = {'': {'': dt.datetime(2009, 1, 1),
                    'all': dt.datetime(2009, 1, 1),
                    'prelim': dt.datetime(2009, 1, 1),
                    'daily': tomorrow,
                    'forecast': tomorrow,
                    '45day': tomorrow}}

# Other tags assumed to be True
_test_download_travis = {'': {'prelim': False}}

# ----------------------------------------------------------------------------
# Instrument methods


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    """

    self.acknowledgements = mm_f107.acknowledgements(self.name, self.tag)
    self.references = mm_f107.references(self.name, self.tag)
    logger.info(self.acknowledgements)
    return


def clean(self):
    """ Cleaning function for Space Weather indices

    Note
    ----
    F10.7 doesn't require cleaning
    """
    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag=None, inst_id=None):
    """Load F10.7 index files

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames
    tag : str or NoneType
        tag or None (default=None)
    inst_id : str or NoneType
        satellite id or None (default=None)

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data
    meta : pysat.Meta
        Object containing metadata such as column names and units

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """
    all_file_results = []
    for fname in fnames:
        if tag == '':
            # F10.7 data is stored monthly an pysat needs daily data.  The
            # daily date is attached to filename.  Parse off the last date,
            # load month of data, downselect to desired day
            date = dt.datetime.strptime(fname[-10:], '%Y-%m-%d')
            data = pds.read_csv(fname[0:-11], index_col=0, parse_dates=True)
            idx, = np.where((data.index >= date)
                            & (data.index < date + pds.DateOffset(days=1)))
            result = data.iloc[idx, :]
        elif tag in ['all', 'daily', 'prelim', 'forecast', '45day']:
            result = pds.read_csv(fname, index_col=0, parse_dates=True)
        all_file_results.append(result)
    # combine loaded data together
    result = pds.concat(all_file_results, axis=0, sort=True)

    meta = pysat.Meta()
    meta['f107'] = {meta.labels.units: 'SFU',
                    meta.labels.name: 'F10.7 cm solar index',
                    meta.labels.desc:
                    'F10.7 cm radio flux in Solar Flux Units (SFU)'}

    if tag == '45day':
        meta['ap'] = {meta.labels.name: 'Daily Ap index',
                      meta.labels.desc: 'Daily average of 3-h ap indices'}
    elif tag == 'daily' or tag == 'prelim':
        meta['ssn'] = {meta.labels.name: 'Sunspot Number',
                       meta.labels.desc: 'SESC Sunspot Number',
                       meta.labels.fill_val: -999}
        meta['ss_area'] = {meta.labels.name: 'Sunspot Area',
                           meta.labels.desc: 'Sunspot Area 10$^6$ Hemisphere',
                           meta.labels.fill_val: -999}
        meta['new_reg'] = {meta.labels.name: 'New Regions',
                           meta.labels.desc: 'New active solar regions',
                           meta.labels.fill_val: -999}
        meta['smf'] = {meta.labels.name: 'Solar Mean Field',
                       meta.labels.desc: 'Standford Solar Mean Field',
                       meta.labels.fill_val: -999}
        meta['goes_bgd_flux'] = {meta.labels.name: 'X-ray Background Flux',
                                 meta.labels.desc:
                                 'GOES15 X-ray Background Flux',
                                 meta.labels.fill_val: '*'}
        meta['c_flare'] = {meta.labels.name: 'C X-Ray Flares',
                           meta.labels.desc: 'C-class X-Ray Flares',
                           meta.labels.fill_val: -1}
        meta['m_flare'] = {meta.labels.name: 'M X-Ray Flares',
                           meta.labels.desc: 'M-class X-Ray Flares',
                           meta.labels.fill_val: -1}
        meta['x_flare'] = {meta.labels.name: 'X X-Ray Flares',
                           meta.labels.desc: 'X-class X-Ray Flares',
                           meta.labels.fill_val: -1}
        meta['o1_flare'] = {meta.labels.name: '1 Optical Flares',
                            meta.labels.desc: '1-class Optical Flares',
                            meta.labels.fill_val: -1}
        meta['o2_flare'] = {meta.labels.name: '2 Optical Flares',
                            meta.labels.desc: '2-class Optical Flares',
                            meta.labels.fill_val: -1}
        meta['o3_flare'] = {meta.labels.name: '3 Optical Flares',
                            meta.labels.desc: '3-class Optical Flares',
                            meta.labels.fill_val: -1}

    return result, meta


def list_files(tag=None, inst_id=None, data_path=None, format_str=None):
    """Return a Pandas Series of every file for F10.7 data

    Parameters
    ----------
    tag : string or NoneType
        Denotes type of file to load.
        (default=None)
    inst_id : string or NoneType
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    data_path : string or NoneType
        Path to data directory.  If None is specified, the value previously
        set in Instrument.files.data_path is used.  (default=None)
    format_str : string or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    pysat.Files.from_os : pysat._files.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    if data_path is not None:
        if tag == '':
            # files are by month, going to add date to monthly filename for
            # each day of the month. The load routine will load a month of
            # data and use the appended date to select out appropriate data.
            if format_str is None:
                format_str = 'f107_monthly_{year:04d}-{month:02d}.txt'
            out = pysat.Files.from_os(data_path=data_path,
                                      format_str=format_str)
            if not out.empty:
                out.loc[out.index[-1] + pds.DateOffset(months=1)
                        - pds.DateOffset(days=1)] = out.iloc[-1]
                out = out.asfreq('D', 'pad')
                out = out + '_' + out.index.strftime('%Y-%m-%d')
            return out

        elif tag == 'all':
            # files are by year
            if format_str is None:
                format_str = ''.join(('f107_1947_to_{year:04d}-{month:02d}-',
                                      '{day:02d}.txt'))
            out = pysat.Files.from_os(data_path=data_path,
                                      format_str=format_str)
            # load the same data (all), regardless of which day a user selects
            # resample file list to provide the same filename for every day
            # of f107 data
            if not out.empty:
                # only want to use the most recent file
                out = out.iloc[-1:]
                # first day of data is 2-14, ensure same file for first and
                # most recent day
                out.loc[dt.datetime(1947, 2, 13)] = out.iloc[0]
                # make sure things are in order and copy latest filename for
                # all days, thus no matter which day with data the user loads
                # they get the most recent F10.7 file
                out = out.sort_index()
                out = out.asfreq('D', 'pad')

            return out

        elif tag == 'prelim':
            # files are by year (and quarter). The load routine will load a
            # year of data
            if format_str is None:
                format_str = \
                    'f107_prelim_{year:04d}_{month:02d}_v{version:01d}.txt'
            out = pysat.Files.from_os(data_path=data_path,
                                      format_str=format_str)

            if not out.empty:
                # Set each file's valid length at a 1-day resolution
                orig_files = out.sort_index().copy()
                new_files = list()

                for orig in orig_files.iteritems():
                    # Version determines each file's valid length
                    version = int(orig[1].split("_v")[1][0])
                    doff = pds.DateOffset(years=1) if version == 2 \
                        else pds.DateOffset(months=3)
                    istart = orig[0]
                    iend = istart + doff - pds.DateOffset(days=1)

                    # Ensure the end time does not extend past the number of
                    # possible days included based on the file's download time
                    fname = os.path.join(data_path, orig[1])
                    dend = dt.datetime.utcfromtimestamp(os.path.getctime(fname))
                    dend = dend - pds.DateOffset(days=1)
                    if dend < iend:
                        iend = dend

                    # Pad the original file index
                    out.loc[iend] = orig[1]
                    out = out.sort_index()

                    # Save the files at a daily cadence over the desired period
                    new_files.append(out.loc[istart: iend].asfreq('D', 'pad'))
                # Add the newly indexed files to the file output
                out = pds.concat(new_files, sort=True)
                out = out.dropna()
                out = out.sort_index()

            return out

        elif tag == 'daily':
            format_str = 'f107_daily_{year:04d}-{month:02d}-{day:02d}.txt'
            files = pysat.Files.from_os(data_path=data_path,
                                        format_str=format_str)

            # pad list of files data to include most recent file under tomorrow
            if not files.empty:
                pds_off = pds.DateOffset(days=1)
                files.loc[files.index[-1] + pds_off] = files.values[-1]
                files.loc[files.index[-1] + pds_off] = files.values[-1]
            return files

        elif tag == 'forecast':
            format_str = 'f107_forecast_{year:04d}-{month:02d}-{day:02d}.txt'
            files = pysat.Files.from_os(data_path=data_path,
                                        format_str=format_str)
            # pad list of files data to include most recent file under tomorrow
            if not files.empty:
                pds_off = pds.DateOffset(days=1)
                files.loc[files.index[-1] + pds_off] = files.values[-1]
                files.loc[files.index[-1] + pds_off] = files.values[-1]
            return files

        elif tag == '45day':
            format_str = 'f107_45day_{year:04d}-{month:02d}-{day:02d}.txt'
            files = pysat.Files.from_os(data_path=data_path,
                                        format_str=format_str)

            # pad list of files data to include most recent file under tomorrow
            if not files.empty:
                pds_off = pds.DateOffset(days=1)
                files.loc[files.index[-1] + pds_off] = files.values[-1]
                files.loc[files.index[-1] + pds_off] = files.values[-1]
            return files

        else:
            raise ValueError(' '.join(('Unrecognized tag name for Space',
                                       'Weather Index F107')))
    else:
        raise ValueError(' '.join(('A data_path must be passed to the loading',
                                   'routine for F107')))


def download(date_array, tag, inst_id, data_path):
    """Routine to download F107 index data

    Parameters
    -----------
    tag : string or NoneType
        Denotes type of file to load.  Accepted types are '' and 'forecast'.
        (default=None)
    inst_id : string or NoneType
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    data_path : string or NoneType
        Path to data directory.  If None is specified, the value previously
        set in Instrument.files.data_path is used.  (default=None)

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    Only able to download current forecast data, not archived forecasts.

    """

    # download standard F107 data
    if tag == '':
        # download from LASP, by month
        for date in date_array:
            # modify date to be the start of the month
            if date.day != 1:
                raise ValueError(' '.join(('The Download routine must be',
                                           'invoked with a freq="MS"',
                                           'option.')))
            # download webpage
            dstr = 'http://lasp.colorado.edu/lisird/latis/dap/'
            dstr += 'noaa_radio_flux.json?time%3E='
            dstr += date.strftime('%Y-%m-%d')
            dstr += 'T00:00:00.000Z&time%3C='
            dstr += (date + pds.DateOffset(months=1)
                     - pds.DateOffset(days=1)).strftime('%Y-%m-%d')
            dstr += 'T00:00:00.000Z'
            # data returned as json
            r = requests.get(dstr)
            # process
            raw_dict = json.loads(r.text)['noaa_radio_flux']
            data = pds.DataFrame.from_dict(raw_dict['samples'])
            if data.empty:
                warnings.warn("no data for {:}".format(date), UserWarning)
            else:
                times = [dt.datetime.strptime(time, '%Y%m%d')
                         for time in data.pop('time')]
                data.index = times
                # replace fill with NaNs
                idx, = np.where(data['f107'] == -99999.0)
                data.iloc[idx, :] = np.nan
                # create file
                str_date = date.strftime('%Y-%m')
                data_file = 'f107_monthly_{:s}.txt'.format(str_date)
                data.to_csv(os.path.join(data_path, data_file), header=True)

    elif tag == 'all':
        # download from LASP, by year

        # download webpage
        dstr = 'http://lasp.colorado.edu/lisird/latis/dap/'
        dstr += 'noaa_radio_flux.json?time%3E='
        dstr += dt.datetime(1947, 2, 13).strftime('%Y-%m-%d')
        dstr += 'T00:00:00.000Z&time%3C='
        now = dt.datetime.utcnow()
        dstr += now.strftime('%Y-%m-%dT%H:%M:%S.000Z')
        # data returned as json
        r = requests.get(dstr)
        # process
        raw_dict = json.loads(r.text)['noaa_radio_flux']
        data = pds.DataFrame.from_dict(raw_dict['samples'])
        try:
            # This is the new data format
            times = [dt.datetime.strptime(time, '%Y%m%d')
                     for time in data.pop('time')]
        except ValueError:
            # Accepts old file formats
            times = [dt.datetime.strptime(time, '%Y %m %d')
                     for time in data.pop('time')]
        data.index = times
        # replace fill with NaNs
        idx, = np.where(data['f107'] == -99999.0)
        data.iloc[idx, :] = np.nan
        # create file
        data_file = 'f107_1947_to_{:s}.txt'.format(now.strftime('%Y-%m-%d'))
        data.to_csv(os.path.join(data_path, data_file), header=True)

    elif tag == 'prelim':
        ftp = ftplib.FTP('ftp.swpc.noaa.gov')  # connect to host, default port
        ftp.login()  # user anonymous, passwd anonymous@
        ftp.cwd('/pub/indices/old_indices')

        bad_fname = list()

        # Get the local files, to ensure that the version 1 files are
        # downloaded again if more data has been added
        local_files = list_files(tag, inst_id, data_path)

        # To avoid downloading multiple files, cycle dates based on file length
        date = date_array[0]
        while date <= date_array[-1]:
            # The file name changes, depending on how recent the requested
            # data is
            qnum = (date.month - 1) // 3 + 1  # Integer floor division
            qmonth = (qnum - 1) * 3 + 1
            quar = 'Q{:d}_'.format(qnum)
            fnames = ['{:04d}{:s}DSD.txt'.format(date.year, ss)
                      for ss in ['_', quar]]
            versions = ["01_v2", "{:02d}_v1".format(qmonth)]
            vend = [dt.datetime(date.year, 12, 31),
                    dt.datetime(date.year, qmonth, 1)
                    + pds.DateOffset(months=3) - pds.DateOffset(days=1)]
            downloaded = False
            rewritten = False

            # Attempt the download(s)
            for iname, fname in enumerate(fnames):
                # Test to see if we already tried this filename
                if fname in bad_fname:
                    continue

                local_fname = fname
                saved_fname = os.path.join(data_path, local_fname)
                ofile = '_'.join(['f107', 'prelim', '{:04d}'.format(date.year),
                                  '{:s}.txt'.format(versions[iname])])
                outfile = os.path.join(data_path, ofile)

                if os.path.isfile(outfile):
                    downloaded = True

                    # Check the date to see if this should be rewritten
                    checkfile = os.path.split(outfile)[-1]
                    has_file = local_files == checkfile
                    if np.any(has_file):
                        if has_file[has_file].index[-1] < vend[iname]:
                            # This file will be updated again, but only attempt
                            # to do so if enough time has passed from the
                            # last time it was downloaded
                            yesterday = today - pds.DateOffset(days=1)
                            if has_file[has_file].index[-1] < yesterday:
                                rewritten = True
                else:
                    # The file does not exist, if it can be downloaded, it
                    # should be 'rewritten'
                    rewritten = True

                # Attempt to download if the file does not exist or if the
                # file has been updated
                if rewritten or not downloaded:
                    try:
                        sys.stdout.flush()
                        ftp.retrbinary('RETR ' + fname,
                                       open(saved_fname, 'wb').write)
                        downloaded = True
                        logger.info(' '.join(('Downloaded file for ',
                                              date.strftime('%x'))))

                    except ftplib.error_perm as exception:
                        # Could not fetch, so cannot rewrite
                        rewritten = False

                        # Test for an error
                        if str(exception.args[0]).split(" ", 1)[0] != '550':
                            raise RuntimeError(exception)
                        else:
                            # file isn't actually there, try the next name
                            os.remove(saved_fname)

                            # Save this so we don't try again
                            # Because there are two possible filenames for
                            # each time, it's ok if one isn't there.  We just
                            # don't want to keep looking for it.
                            bad_fname.append(fname)

                # If the first file worked, don't try again
                if downloaded:
                    break

            if not downloaded:
                logger.info(' '.join(('File not available for',
                                      date.strftime('%x'))))
            elif rewritten:
                with open(saved_fname, 'r') as fprelim:
                    lines = fprelim.read()

                rewrite_daily_file(date.year, outfile, lines)
                os.remove(saved_fname)

            # Cycle to the next date
            date = vend[iname] + pds.DateOffset(days=1)

        # Close connection after downloading all dates
        ftp.close()

    elif tag == 'daily':
        logger.info('This routine can only download the latest 30 day file')

        # download webpage
        furl = 'https://services.swpc.noaa.gov/text/daily-solar-indices.txt'
        r = requests.get(furl)

        # Save the output
        data_file = 'f107_daily_{:s}.txt'.format(today.strftime('%Y-%m-%d'))
        outfile = os.path.join(data_path, data_file)
        rewrite_daily_file(today.year, outfile, r.text)

    elif tag == 'forecast':
        logger.info(' '.join(('This routine can only download the current',
                              'forecast, not archived forecasts')))
        # download webpage
        furl = ''.join(('https://services.swpc.noaa.gov/text/',
                        '3-day-solar-geomag-predictions.txt'))
        r = requests.get(furl)
        # parse text to get the date the prediction was generated
        date_str = r.text.split(':Issued: ')[-1].split(' UTC')[0]
        date = dt.datetime.strptime(date_str, '%Y %b %d %H%M')
        # get starting date of the forecasts
        raw_data = r.text.split(':Prediction_dates:')[-1]
        forecast_date = dt.datetime.strptime(raw_data[3:14], '%Y %b %d')
        # times for output data
        times = pds.date_range(forecast_date, periods=3, freq='1D')
        # string data is the forecast value for the next three days
        raw_data = r.text.split('10cm_flux:')[-1]
        raw_data = raw_data.split('\n')[1]
        val1 = int(raw_data[24:27])
        val2 = int(raw_data[38:41])
        val3 = int(raw_data[52:])

        # put data into nicer DataFrame
        data = pds.DataFrame([val1, val2, val3], index=times, columns=['f107'])
        # write out as a file
        data_file = 'f107_forecast_{:s}.txt'.format(date.strftime('%Y-%m-%d'))
        data.to_csv(os.path.join(data_path, data_file), header=True)

    elif tag == '45day':
        logger.info(' '.join(('This routine can only download the current',
                              'forecast, not archived forecasts')))
        # download webpage
        furl = 'https://services.swpc.noaa.gov/text/45-day-ap-forecast.txt'
        r = requests.get(furl)
        # parse text to get the date the prediction was generated
        date_str = r.text.split(':Issued: ')[-1].split(' UTC')[0]
        date = dt.datetime.strptime(date_str, '%Y %b %d %H%M')
        # get to the forecast data
        raw_data = r.text.split('45-DAY AP FORECAST')[-1]
        # grab AP part
        raw_ap = raw_data.split('45-DAY F10.7 CM FLUX FORECAST')[0]
        # clean up
        raw_ap = raw_ap.split('\n')[1:-1]
        # f107
        raw_f107 = raw_data.split('45-DAY F10.7 CM FLUX FORECAST')[-1]
        # clean up
        raw_f107 = raw_f107.split('\n')[1:-4]

        # parse the AP data
        ap_times, ap = parse_45day_block(raw_ap)

        # parse the F10.7 data
        f107_times, f107 = parse_45day_block(raw_f107)

        # collect into DataFrame
        data = pds.DataFrame(f107, index=f107_times, columns=['f107'])
        data['ap'] = ap
        # write out as a file
        data_file = 'f107_45day_{:s}.txt'.format(date.strftime('%Y-%m-%d'))
        data.to_csv(os.path.join(data_path, data_file), header=True)

    return


# ----------------------------------------------------------------------------
# Local functions


def parse_45day_block(block_lines):
    """ Parse the data blocks used in the 45-day Ap and F10.7 Flux Forecast
    file

    Parameters
    ----------
    block_lines : list
        List of lines containing data in this data block

    Returns
    -------
    dates : list
        List of dates for each date/data pair in this block
    values : list
        List of values for each date/data pair in this block

    """

    # Initialize the output
    dates = list()
    values = list()

    # Cycle through each line in this block
    for line in block_lines:
        # Split the line on whitespace
        split_line = line.split()

        # Format the dates
        dates.extend([dt.datetime.strptime(tt, "%d%b%y")
                      for tt in split_line[::2]])

        # Format the data values
        values.extend([int(vv) for vv in split_line[1::2]])

    return dates, values


def rewrite_daily_file(year, outfile, lines):
    """ Rewrite the SWPC Daily Solar Data files

    Parameters
    ----------
    year : int
        Year of data file (format changes based on date)
    outfile : str
        Output filename
    lines : str
        String containing all output data (result of 'read')

    """

    # get to the solar index data
    if year > 2000:
        raw_data = lines.split('#---------------------------------')[-1]
        raw_data = raw_data.split('\n')[1:-1]
        optical = True
    else:
        raw_data = lines.split('# ')[-1]
        raw_data = raw_data.split('\n')
        optical = False if raw_data[0].find('Not Available') or year == 1994 \
            else True
        istart = 7 if year < 2000 else 1
        raw_data = raw_data[istart:-1]

    # parse the data
    solar_times, data_dict = parse_daily_solar_data(raw_data, year, optical)

    # collect into DataFrame
    data = pds.DataFrame(data_dict, index=solar_times,
                         columns=data_dict.keys())

    # write out as a file
    data.to_csv(outfile, header=True)

    return


def parse_daily_solar_data(data_lines, year, optical):
    """ Parse the data in the SWPC daily solar index file

    Parameters
    ----------
    data_lines : list
        List of lines containing data
    year : list
        Year of file
    optical : boolean
        Flag denoting whether or not optical data is available

    Returns
    -------
    dates : list
        List of dates for each date/data pair in this block
    values : dict
        Dict of lists of values, where each key is the value name

    """

    # Initialize the output
    dates = list()
    val_keys = ['f107', 'ssn', 'ss_area', 'new_reg', 'smf', 'goes_bgd_flux',
                'c_flare', 'm_flare', 'x_flare', 'o1_flare', 'o2_flare',
                'o3_flare']
    optical_keys = ['o1_flare', 'o2_flare', 'o3_flare']
    xray_keys = ['c_flare', 'm_flare', 'x_flare']
    values = {kk: list() for kk in val_keys}

    # Cycle through each line in this file
    for line in data_lines:
        # Split the line on whitespace
        split_line = line.split()

        # Format the date
        dfmt = "%Y %m %d" if year > 1996 else "%d %b %y"
        dates.append(dt.datetime.strptime(" ".join(split_line[0:3]), dfmt))

        # Format the data values
        j = 0
        for i, kk in enumerate(val_keys):
            if year == 1994 and kk == 'new_reg':
                # New regions only in files after 1994
                val = -999
            elif((year == 1994 and kk in xray_keys)
                 or (not optical and kk in optical_keys)):
                # X-ray flares in files after 1994, optical flares come later
                val = -1
            else:
                val = split_line[j + 3]
                j += 1

            if kk != 'goes_bgd_flux':
                if val == "*":
                    val = -999 if i < 5 else -1
                else:
                    val = int(val)
            values[kk].append(val)

    return dates, values


def calc_f107a(f107_inst, f107_name='f107', f107a_name='f107a', min_pnts=41):
    """ Calculate the 81 day mean F10.7

    Parameters
    ----------
    f107_inst : pysat.Instrument
        pysat Instrument holding the F10.7 data
    f107_name : str
        Data column name for the F10.7 data (default='f107')
    f107a_name : str
        Data column name for the F10.7a data (default='f107a')
    min_pnts : int
        Minimum number of points required to calculate an average (default=41)

    Returns
    -------
    Void : Updates f107_inst with F10.7a data

    Notes
    -----
    Will not pad data on its own

    """

    # Test to see that the input data is present
    if f107_name not in f107_inst.data.columns:
        raise ValueError("unknown input data column: " + f107_name)

    # Test to see that the output data does not already exist
    if f107a_name in f107_inst.data.columns:
        raise ValueError("output data column already exists: " + f107a_name)

    if f107_name in f107_inst.meta:
        fill_val = f107_inst.meta[f107_name][f107_inst.meta.labels.fill_val]
    else:
        fill_val = np.nan

    # Calculate the rolling mean.  Since these values are centered but rolling
    # function doesn't allow temporal windows to be calculated this way, create
    # a hack for this.
    #
    # Ensure the data are evenly sampled at a daily frequency, since this is
    # how often F10.7 is calculated.
    f107_fill = f107_inst.data.resample('1D').fillna(method=None)

    # Replace the time index with an ordinal
    time_ind = f107_fill.index
    f107_fill['ord'] = pds.Series([tt.toordinal() for tt in time_ind],
                                  index=time_ind)
    f107_fill.set_index('ord', inplace=True)

    # Calculate the mean
    f107_fill[f107a_name] = f107_fill[f107_name].rolling(window=81,
                                                         min_periods=min_pnts,
                                                         center=True).mean()

    # Replace the ordinal index with the time
    f107_fill['time'] = pds.Series(time_ind, index=f107_fill.index)
    f107_fill.set_index('time', inplace=True)

    # Resample to the original frequency, if it is not equal to 1 day
    freq = pysat.utils.time.calc_freq(f107_inst.index)
    if freq != "86400S":
        # Resample to the desired frequency
        f107_fill = f107_fill.resample(freq).pad()

        # Save the output in a list
        f107a = list(f107_fill[f107a_name])

        # Fill any dates that fall
        time_ind = pds.date_range(f107_fill.index[0], f107_inst.index[-1],
                                  freq=freq)
        for itime in time_ind[f107_fill.index.shape[0]:]:
            if (itime - f107_fill.index[-1]).total_seconds() < 86400.0:
                f107a.append(f107a[-1])
            else:
                f107a.append(fill_val)

        # Redefine the Series
        f107_fill = pds.DataFrame({f107a_name: f107a}, index=time_ind)

    # There may be missing days in the output data, remove these
    if f107_inst.index.shape < f107_fill.index.shape:
        f107_fill = f107_fill.loc[f107_inst.index]

    # Save the data
    f107_inst[f107a_name] = f107_fill[f107a_name]

    # Update the metadata
    meta_dict = {f107_inst.meta.labels.units: 'SFU',
                 f107_inst.meta.labels.name: 'F10.7a',
                 f107_inst.meta.labels.desc: "81-day centered average of F10.7",
                 f107_inst.meta.labels.min_val: 0.0,
                 f107_inst.meta.labels.max_val: np.nan,
                 f107_inst.meta.labels.fill_val: fill_val,
                 f107_inst.meta.labels.notes:
                 ' '.join(('Calculated using data between',
                           '{:} and {:}'.format(f107_inst.index[0],
                                                f107_inst.index[-1])))}

    f107_inst.meta[f107a_name] = meta_dict

    return



from pysatSpaceWeather.instruments import methods  # noqa F401

__all__ = ['sw_dst', 'sw_f107', 'sw_kp', 'sw_ace']

for inst in __all__:
    exec("from pysatSpaceWeather.instruments import {x}".format(x=inst))



import os

from pysatSpaceWeather import instruments  # noqa F401

# Set directory for test data
here = os.path.abspath(os.path.dirname(__file__))
test_data_path = os.path.join(here, 'tests', 'test_data')



# -*- coding: utf-8 -*-
"""Supports ACE Magnetometer data

Properties
----------
platform
    'sw'
name
    'ace'
tag
    - 'realtime' Real-time data from the Space Weather Prediction Center (SWPC)
    - 'historic' Historic data from the SWPC
inst_id
    - 'mag' Magnetometer
    - 'swepam' Solar Wind Electron Proton Alpha Monitor
    - 'sis' Solar Isotope Spectrometer
    - 'epam' Electron, Proton, and Alpha Monitor

Note
----
The real-time data is stored by generation date, where each file contains the
data for the current day.  If you leave download dates empty, though, it will
grab today's file three times and assign dates from yesterday, today, and
tomorrow.
::

    mag = pysat.Instrument('sw', 'ace', inst_id='mag', tag='realtime')
    mag.download(start=dt.datetime.now(), stop=dt.datetime.now())
    mag.load(date=dt.datetime.now())



Warnings
--------
The real-time data should not be used with the data padding option available
from pysat.Instrument objects.

"""

import datetime as dt
import numpy as np
import os
import pandas as pds
import requests

import pysat

from pysatSpaceWeather.instruments.methods import ace as mm_ace

logger = pysat.logger

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'ace'
tags = {'realtime': 'Real-time data from the SWPC',
        'historic': ' Historic data from the SWPC'}
inst_ids = {inst_id: [tag for tag in tags.keys()]
            for inst_id in ['mag', 'swepam', 'epam', 'sis']}

# Define today's date
now = dt.datetime.now()

# ----------------------------------------------------------------------------
# Instrument test attributes

# Set test dates (first level: inst_id, second level: tag)
_test_dates = {inst_id: {'realtime': dt.datetime(now.year, now.month, now.day),
                         'historic': dt.datetime(2009, 1, 1)}
               for inst_id in inst_ids.keys()}

# ----------------------------------------------------------------------------
# Instrument methods


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    """

    # Set the appropraite acknowledgements and references
    self.acknowledgements = mm_ace.acknowledgements(self.platform)
    self.references = mm_ace.references(self.inst_id)

    logger.info(self.acknowledgements)

    return


def clean(self):
    """Routine to clean real-time ACE data using the status flag

    Note
    ----
    Supports 'clean' and 'dirty'.  Replaces all fill values with NaN.
    Clean - status flag of zero (nominal data)
    Dirty - status flag < 9 (accepts bad data record, removes no data record)

    """

    # Set the clean level
    if self.clean_level == 'dusty':
        logger.warning("unused clean level 'dusty', reverting to 'clean'")
        self.clean_level = 'clean'

    max_status = 9
    if self.clean_level == "clean":
        max_status = 0
    elif self.clean_level == "dirty":
        max_status = 8

    # Replace all fill values with NaN
    for col in self.data.columns:
        fill_val = self.meta[col][self.meta.labels.fill_val]
        if ~np.isnan(fill_val):
            self.data[col] = self.data[col].replace(fill_val, np.nan)

    # Replace bad values with NaN and remove times with no valid data
    if self.inst_id in ['mag', 'swepam']:
        self.data = self.data[self.data['status'] <= max_status]
    else:
        if self.inst_id == 'epam':
            ecols = ['eflux_38-53', 'eflux_175-315']
            # Evaluate the electron flux data
            for col in ecols:
                self.data[col][self.data['status_e'] > max_status] = np.nan

            # Evaluate the proton flux data
            pcols = ['pflux_47-68', 'pflux_115-195', 'pflux_310-580',
                     'pflux_795-1193', 'pflux_1060-1900']
            for col in pcols:
                self.data[col][self.data['status_p'] > max_status] = np.nan

            # Include both fluxes and the anisotropy index in the removal eval
            eval_cols = ecols + pcols
            eval_cols.append('anis_ind')
        elif self.inst_id == 'sis':
            # Evaluate the different proton fluxes
            self.data['int_pflux_10MeV'][
                self.data['status_10'] > max_status] = np.nan
            self.data['int_pflux_30MeV'][
                self.data['status_30'] > max_status] = np.nan

            eval_cols = ['int_pflux_10MeV', 'int_pflux_30MeV']
        else:
            raise ValueError('unknown inst_id {:}'.format(self.inst_id))

        # Remove lines without any good data
        good_cols = (~np.isnan(self.data.loc[:, eval_cols])).sum(axis=1)
        bad_index = good_cols[good_cols == 0].index
        self.data = self.data.drop(index=bad_index)

    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag=None, inst_id=None):
    """Load the ACE space weather prediction data

    Parameters
    ----------
    fnames : array-like
        Series, list, or array of filenames
    tag : str or NoneType
        tag or None (default=None)
    inst_id : str or NoneType
        ACE instrument or None (default=None)

    Returns
    -------
    data : pandas.DataFrame
        Object containing instrument data
    meta : pysat.Meta
        Object containing metadata such as column names and units

    Raises
    ------
    ValueError
        When unknown inst_id is supplied.

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    # Save each file to the output DataFrame
    data = pds.DataFrame()
    for fname in fnames:
        result = pds.read_csv(fname, index_col=0, parse_dates=True)
        data = pds.concat([data, result], axis=0)

    # Assign the meta data
    meta = pysat.Meta()
    meta['jd'] = {meta.labels.units: 'days',
                  meta.labels.name: 'MJD',
                  meta.labels.desc: 'Modified Julian Day'}
    meta['sec'] = {meta.labels.units: 's',
                   meta.labels.name: 'Sec of Day',
                   meta.labels.desc: 'Seconds of Julian Day'}
    status_desc = '0 = nominal data, 1 to 8 = bad data record, 9 = no data'

    if inst_id == 'mag':
        meta['status'] = {meta.labels.name: 'Status',
                          meta.labels.desc: status_desc}
        meta['bx_gsm'] = {meta.labels.units: 'nT',
                          meta.labels.name: 'Bx GSM',
                          meta.labels.desc: '1-min averaged IMF Bx',
                          meta.labels.fill_val: -999.9}
        meta['by_gsm'] = {meta.labels.units: 'nT',
                          meta.labels.name: 'By GSM',
                          meta.labels.desc: '1-min averaged IMF By',
                          meta.labels.fill_val: -999.9}
        meta['bz_gsm'] = {meta.labels.units: 'nT',
                          meta.labels.name: 'Bz GSM',
                          meta.labels.desc: '1-min averaged IMF Bz',
                          meta.labels.fill_val: -999.9}
        meta['bt_gsm'] = {meta.labels.units: 'nT',
                          meta.labels.name: 'Bt GSM',
                          meta.labels.desc: '1-min averaged IMF Bt',
                          meta.labels.fill_val: -999.9}
        meta['lat_gsm'] = {meta.labels.units: 'degrees',
                           meta.labels.name: 'GSM Lat',
                           meta.labels.desc: 'GSM Latitude',
                           meta.labels.fill_val: -999.9,
                           meta.labels.min_val: -90.0,
                           meta.labels.max_val: 90.0}
        meta['lon_gsm'] = {meta.labels.units: 'degrees',
                           meta.labels.name: 'GSM Lon',
                           meta.labels.desc: 'GSM Longitude',
                           meta.labels.fill_val: -999.9,
                           meta.labels.min_val: 0.0, meta.labels.max_val: 360.0}
    elif inst_id == 'epam':
        flux_desc = '5-min averaged Differential '
        meta['status_e'] = {meta.labels.name: 'Diff e- Flux Status',
                            meta.labels.desc: status_desc}
        meta['status_p'] = {meta.labels.name: 'Diff Proton Flux Status',
                            meta.labels.desc: status_desc}
        meta['anis_ind'] = {meta.labels.name: 'Anisotropy Index',
                            meta.labels.desc: 'Range: 0.0 - 2.0',
                            meta.labels.fill_val: -1.0}
        meta['eflux_38-53'] = {meta.labels.units: 'particles/cm2-s-ster-MeV',
                               meta.labels.name: 'Diff e- Flux 38-53 eV',
                               meta.labels.desc:
                               ''.join([flux_desc,
                                        'Electron Flux between 35-53 eV']),
                               meta.labels.fill_val: -1.0e5}
        meta['eflux_175-315'] = {meta.labels.units: 'particles/cm2-s-ster-MeV',
                                 meta.labels.name: 'Diff e- Flux 175-315 eV',
                                 meta.labels.desc:
                                 ''.join([flux_desc,
                                          'Electron Flux between 175-315 eV']),
                                 meta.labels.fill_val: -1.0e5}
        meta['pflux_47-68'] = {meta.labels.units: 'particles/cm2-s-ster-MeV',
                               meta.labels.name: 'Diff Proton Flux 47-68 keV',
                               meta.labels.desc:
                               ''.join([flux_desc,
                                        'Proton Flux between 47-68 keV']),
                               meta.labels.fill_val: -1.0e5}
        meta['pflux_115-195'] = {meta.labels.units: 'particles/cm2-s-ster-MeV',
                                 meta.labels.name:
                                 'Diff Proton Flux 115-195 keV',
                                 meta.labels.desc:
                                 ''.join([flux_desc,
                                          'Proton Flux between 115-195 keV']),
                                 meta.labels.fill_val: -1.0e5}
        meta['pflux_310-580'] = {meta.labels.units: 'particles/cm2-s-ster-MeV',
                                 meta.labels.name:
                                 'Diff Proton Flux 310-580 keV',
                                 meta.labels.desc:
                                 ''.join([flux_desc,
                                          'Proton Flux between 310-580 keV']),
                                 meta.labels.fill_val: -1.0e5}
        meta['pflux_795-1193'] = {meta.labels.units: 'particles/cm2-s-ster-MeV',
                                  meta.labels.name:
                                  'Diff Proton Flux 795-1193 keV',
                                  meta.labels.desc:
                                  ''.join([flux_desc,
                                           'Proton Flux between 795-1193 ',
                                           'keV']),
                                  meta.labels.fill_val: -1.0e5}
        meta['pflux_1060-1900'] = {meta.labels.units:
                                   'particles/cm2-s-ster-MeV',
                                   meta.labels.name:
                                   'Diff Proton Flux 1060-1900 keV',
                                   meta.labels.desc:
                                   ''.join([flux_desc,
                                            'Proton Flux between 1060-1900',
                                            ' keV']),
                                   meta.labels.fill_val: -1.0e5}
    elif inst_id == "swepam":
        sw_desc = '1-min averaged Solar Wind '
        meta['status'] = {meta.labels.name: 'Status',
                          meta.labels.desc: status_desc}
        meta['sw_proton_dens'] = {meta.labels.units: 'p/cc',
                                  meta.labels.name: 'Solar Wind Proton Density',
                                  meta.labels.desc: ''.join([sw_desc,
                                                            'Proton Density']),
                                  meta.labels.fill_val: -9999.9}
        meta['sw_bulk_speed'] = {meta.labels.units: 'km/s',
                                 meta.labels.name: 'Solar Wind Bulk Speed',
                                 meta.labels.desc: ''.join([sw_desc,
                                                           'Bulk Speed']),
                                 meta.labels.fill_val: -9999.9}
        meta['sw_ion_temp'] = {meta.labels.units: 'K',
                               meta.labels.name: 'Solar Wind Ti',
                               meta.labels.desc: ''.join([sw_desc,
                                                         'Ion Temperature']),
                               meta.labels.fill_val: -1.0e5}
    elif inst_id == 'sis':
        flux_name = 'Integral Proton Flux'
        meta['status_10'] = {meta.labels.name: ''.join([flux_name,
                                                       ' > 10 MeV Status']),
                             meta.labels.desc: status_desc}
        meta['status_30'] = {meta.labels.name: ''.join([flux_name,
                                                       ' > 30 MeV Status']),
                             meta.labels.desc: status_desc}
        meta['int_pflux_10MeV'] = {meta.labels.units: 'p/cs2-sec-ster',
                                   meta.labels.name: ''.join([flux_name,
                                                             ' > 10 MeV']),
                                   meta.labels.desc: ''.join(['5-min averaged ',
                                                             flux_name,
                                                             ' > 10 MeV']),
                                   meta.labels.fill_val: -1.0e5}
        meta['int_pflux_30MeV'] = {meta.labels.units: 'p/cs2-sec-ster',
                                   meta.labels.name: ''.join([flux_name,
                                                             ' > 30 MeV']),
                                   meta.labels.desc: ''.join(['5-min averaged ',
                                                             flux_name,
                                                             ' > 30 MeV']),
                                   meta.labels.fill_val: -1.0e5}
    else:
        raise ValueError('unknown inst_id {:}'.format(inst_id))

    return data, meta


def list_files(tag=None, inst_id=None, data_path=None, format_str=None):
    """Return a Pandas Series of every file for ACE data

    Parameters
    ----------
    tag : string or NoneType
        Denotes type of file to load.
        (default=None)
    inst_id : string or NoneType
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    data_path : string or NoneType
        Path to data directory.  If None is specified, the value previously
        set in Instrument.files.data_path is used.  (default=None)
    format_str : string or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    pysat.Files.from_os : pysat.utils.files.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    format_str = '_'.join(["ace", inst_id, tag,
                           '{year:04d}-{month:02d}-{day:02d}.txt'])
    if data_path is not None:
        files = pysat.Files.from_os(data_path=data_path, format_str=format_str)
    else:
        raise ValueError(''.join(('A data_path must be passed to the loading',
                                  ' routine for ACE Space Weather data')))

    return files


def download(date_array, tag, inst_id, data_path):
    """Routine to download ACE Space Weather data

    Parameters
    ----------
    date_array : array-like
        Array of datetime values
    tag : string or NoneType
        Denotes type of file to load. Accepted types are 'realtime' and
        'historic' (default=None)
    inst_id : string or NoneType
        Specifies the ACE instrument. Accepts 'mag', 'sis', 'epam', 'swepam'
    data_path : string or NoneType
        Path to data directory. If None is specified, the value previously
        set in Instrument.files.data_path is used (default=None)

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    Only able to download current real-time data

    """

    # Define the file information for each data type and check the
    # date range
    if tag == 'realtime':
        file_fmt = "{:s}-{:s}.txt".format("ace", "magnetometer"
                                          if inst_id == "mag" else inst_id)

        if(len(date_array) > 1 or date_array[0].year != now.year
           or date_array[0].month != now.month or date_array[0] != now.day):
            logger.warning('real-time data only available for current day')
    else:
        data_rate = 1 if inst_id in ['mag', 'swepam'] else 5
        file_fmt = '_'.join(["%Y%m%d", "ace", inst_id,
                             '{:d}m.txt'.format(data_rate)])

    url = {'realtime': 'https://services.swpc.noaa.gov/text/',
           'historic': 'https://sohoftp.nascom.nasa.gov/sdb/ace/daily/'}

    data_cols = {'mag': ['jd', 'sec', 'status', 'bx_gsm', 'by_gsm',
                         'bz_gsm', 'bt_gsm', 'lat_gsm', 'lon_gsm'],
                 "swepam": ['jd', 'sec', 'status', 'sw_proton_dens',
                            'sw_bulk_speed', 'sw_ion_temp'],
                 "epam": ['jd', 'sec', 'status_e', 'eflux_38-53',
                          'eflux_175-315', 'status_p', 'pflux_47-68',
                          'pflux_115-195', 'pflux_310-580',
                          'pflux_795-1193', 'pflux_1060-1900', 'anis_ind'],
                 'sis': ['jd', 'sec', 'status_10', 'int_pflux_10MeV',
                         'status_30', 'int_pflux_30MeV']}

    # Cycle through all the dates
    for dl_date in date_array:
        # download webpage
        furl = ''.join((url[tag], dl_date.strftime(file_fmt)))
        req = requests.get(furl)

        # Split the file at the last header line and then by new line markers
        raw_data = req.text.split('#-----------------')[-1]
        raw_data = raw_data.split('\n')[1:]  # Remove the last header line

        # Parse the file, treating the 4 time columns separately
        data_dict = {col: list() for col in data_cols[inst_id]}
        times = list()
        nsplit = len(data_cols[inst_id]) + 4
        for raw_line in raw_data:
            split_line = raw_line.split()
            if len(split_line) == nsplit:
                times.append(dt.datetime.strptime(' '.join(split_line[:4]),
                                                  '%Y %m %d %H%M'))
                for i, col in enumerate(data_cols[inst_id]):
                    # Convert to a number and save
                    #
                    # Output is saved as a float, so don't bother to
                    # differentiate between int and float
                    data_dict[col].append(float(split_line[4 + i]))
            else:
                if len(split_line) > 0:
                    raise IOError(''.join(['unexpected line encoutered in ',
                                           furl, ":\n", raw_line]))

        # put data into nicer DataFrame
        data = pds.DataFrame(data_dict, index=times)

        # write out as a file
        data_file = '{:s}.txt'.format('_'.join(["ace", inst_id, tag,
                                                dl_date.strftime('%Y-%m-%d')]))
        data.to_csv(os.path.join(data_path, data_file), header=True)

    return



# -*- coding: utf-8 -*-.
"""Provides default routines for solar wind and geospace indices

"""

import pandas as pds
import numpy as np

import pysat

import pysatSpaceWeather as pysat_sw


def acknowledgements(name, tag):
    """Returns acknowledgements for space weather dataset

    Parameters
    ----------
    name : string
        Name of space weather index, eg, dst, f107, kp
    tag : string
        Tag of the space waether index

    """
    lisird = 'NOAA radio flux obtained through LISIRD'
    swpc = ''.join(['Prepared by the U.S. Dept. of Commerce, NOAA, Space ',
                    'Weather Prediction Center'])

    ackn = {'f107': {'': lisird, 'all': lisird, 'prelim': swpc,
                     'daily': swpc, 'forecast': swpc, '45day': swpc}}

    return ackn[name][tag]


def references(name, tag):
    """Returns references for space weather dataset

    Parameters
    ----------
    name : string
        Name of space weather index, eg, dst, f107, kp
    tag : string
        Tag of the space weather index

    """
    noaa_desc = ''.join(['Dataset description: ',
                         'https://www.ngdc.noaa.gov/stp/space-weather/',
                         'solar-data/solar-features/solar-radio/noontime-flux',
                         '/penticton/documentation/dataset-description',
                         '_penticton.pdf, accessed Dec 2020'])
    orig_ref = ''.join(["Covington, A.E. (1948), Solar noise observations on",
                        " 10.7 centimetersSolar noise observations on 10.7 ",
                        "centimeters, Proceedings of the IRE, 36(44), ",
                        "p 454-457."])
    swpc_desc = ''.join(['Dataset description: https://www.swpc.noaa.gov/',
                         'sites/default/files/images/u2/Usr_guide.pdf'])

    refs = {'f107': {'': "\n".join([noaa_desc, orig_ref]),
                     'all': "\n".join([noaa_desc, orig_ref]),
                     'prelim': "\n".join([swpc_desc, orig_ref]),
                     'daily': "\n".join([swpc_desc, orig_ref]),
                     'forecast': "\n".join([swpc_desc, orig_ref]),
                     '45day': "\n".join([swpc_desc, orig_ref])}}

    return refs[name][tag]


def combine_f107(standard_inst, forecast_inst, start=None, stop=None):
    """ Combine the output from the measured and forecasted F10.7 sources

    Parameters
    ----------
    standard_inst : pysat.Instrument or NoneType
        Instrument object containing data for the 'sw' platform, 'f107' name,
        and '', 'all', 'prelim', or 'daily' tag
    forecast_inst : pysat.Instrument or NoneType
        Instrument object containing data for the 'sw' platform, 'f107' name,
        and 'prelim', '45day' or 'forecast' tag
    start : dt.datetime or NoneType
        Starting time for combining data, or None to use earliest loaded
        date from the pysat Instruments (default=None)
    stop : dt.datetime
        Ending time for combining data, or None to use the latest loaded date
        from the pysat Instruments (default=None)

    Returns
    -------
    f107_inst : pysat.Instrument
        Instrument object containing F10.7 observations for the desired period
        of time, merging the standard, 45day, and forecasted values based on
        their reliability

    Notes
    -----
    Merging prioritizes the standard data, then the 45day data, and finally
    the forecast data

    Will not attempt to download any missing data, but will load data

    """

    # Initialize metadata and flags
    notes = "Combines data from"
    stag = standard_inst.tag if len(standard_inst.tag) > 0 else 'default'
    tag = 'combined_{:s}_{:s}'.format(stag, forecast_inst.tag)
    inst_flag = 'standard'

    # If the start or stop times are not defined, get them from the Instruments
    if start is None:
        stimes = [inst.index.min() for inst in [standard_inst, forecast_inst]
                  if len(inst.index) > 0]
        start = min(stimes) if len(stimes) > 0 else None

    if stop is None:
        stimes = [inst.index.max() for inst in [standard_inst, forecast_inst]
                  if len(inst.index) > 0]
        stop = max(stimes) + pds.DateOffset(days=1) if len(stimes) > 0 else None

    if start is None or stop is None:
        raise ValueError(' '.join(("must either load in Instrument objects or",
                                   "provide starting and ending times")))

    if start >= stop:
        raise ValueError("date range is zero or negative")

    # Initialize the output instrument
    f107_inst = pysat.Instrument()
    f107_inst.inst_module = pysat_sw.instruments.sw_f107
    f107_inst.tag = tag
    f107_inst.date = start
    f107_inst.doy = int(start.strftime("%j"))
    fill_val = None

    f107_times = list()
    f107_values = list()

    # Cycle through the desired time range
    itime = start

    while itime < stop and inst_flag is not None:
        # Load and save the standard data for as many times as possible
        if inst_flag == 'standard':
            # Test to see if data loading is needed
            if not np.any(standard_inst.index == itime):
                if standard_inst.tag == 'daily':
                    # Add 30 days
                    standard_inst.load(date=itime + pds.DateOffset(days=30))
                else:
                    standard_inst.load(date=itime)

            good_times = ((standard_inst.index >= itime)
                          & (standard_inst.index < stop))

            if notes.find("standard") < 0:
                notes += " the {:} source ({:} to ".format(inst_flag,
                                                           itime.date())

            if np.any(good_times):
                if fill_val is None:
                    f107_inst.meta = standard_inst.meta
                    fill_val = f107_inst.meta['f107'][
                        f107_inst.meta.labels.fill_val]

                good_vals = standard_inst['f107'][good_times] != fill_val
                new_times = list(standard_inst.index[good_times][good_vals])
                f107_times.extend(new_times)
                new_vals = list(standard_inst['f107'][good_times][good_vals])
                f107_values.extend(new_vals)
                itime = f107_times[-1] + pds.DateOffset(days=1)
            else:
                inst_flag = 'forecast'
                notes += "{:})".format(itime.date())

        # Load and save the forecast data for as many times as possible
        if inst_flag == "forecast":
            # Determine which files should be loaded
            if len(forecast_inst.index) == 0:
                files = np.unique(forecast_inst.files.files[itime:stop])
            else:
                files = [None]  # No load needed, if already initialized

            # Cycle through all possible files of interest, saving relevant
            # data
            for filename in files:
                if filename is not None:
                    forecast_inst.load(fname=filename)

                if notes.find("forecast") < 0:
                    notes += " the {:} source ({:} to ".format(inst_flag,
                                                               itime.date())

                # Check in case there was a problem with the standard data
                if fill_val is None:
                    f107_inst.meta = forecast_inst.meta
                    fill_val = f107_inst.meta['f107'][
                        f107_inst.meta.labels.fill_val]

                # Determine which times to save
                good_times = ((forecast_inst.index >= itime)
                              & (forecast_inst.index < stop))
                good_vals = forecast_inst['f107'][good_times] != fill_val

                # Save desired data and cycle time
                new_times = list(forecast_inst.index[good_times][good_vals])
                f107_times.extend(new_times)
                new_vals = list(forecast_inst['f107'][good_times][good_vals])
                f107_values.extend(new_vals)
                itime = f107_times[-1] + pds.DateOffset(days=1)

            notes += "{:})".format(itime.date())

            inst_flag = None

    if inst_flag is not None:
        notes += "{:})".format(itime.date())

    # Determine if the beginning or end of the time series needs to be padded
    if len(f107_times) >= 2:
        freq = pysat.utils.time.calc_freq(f107_times)
    else:
        freq = None
    end_date = stop - pds.DateOffset(days=1)
    date_range = pds.date_range(start=start, end=end_date, freq=freq)

    if len(f107_times) == 0:
        f107_times = date_range

    date_range = pds.date_range(start=start, end=end_date, freq=freq)

    if date_range[0] < f107_times[0]:
        # Extend the time and value arrays from their beginning with fill
        # values
        itime = abs(date_range - f107_times[0]).argmin()
        f107_times.reverse()
        f107_values.reverse()
        extend_times = list(date_range[:itime])
        extend_times.reverse()
        f107_times.extend(extend_times)
        f107_values.extend([fill_val for kk in extend_times])
        f107_times.reverse()
        f107_values.reverse()

    if date_range[-1] > f107_times[-1]:
        # Extend the time and value arrays from their end with fill values
        itime = abs(date_range - f107_times[-1]).argmin() + 1
        extend_times = list(date_range[itime:])
        f107_times.extend(extend_times)
        f107_values.extend([fill_val for kk in extend_times])

    # Save output data
    f107_inst.data = pds.DataFrame(f107_values, columns=['f107'],
                                   index=f107_times)

    # Resample the output data, filling missing values
    if (date_range.shape != f107_inst.index.shape
            or abs(date_range - f107_inst.index).max().total_seconds() > 0.0):
        f107_inst.data = f107_inst.data.resample(freq).fillna(method=None)
        if np.isfinite(fill_val):
            f107_inst.data[np.isnan(f107_inst.data)] = fill_val

    # Update the metadata notes for this custom procedure
    notes += ", in that order"
    f107_inst.meta['f107'] = {f107_inst.meta.labels.notes: notes}

    return f107_inst



# -*- coding: utf-8 -*-.
"""Provides default routines for solar wind and geospace indices

"""

import pandas as pds
import numpy as np

import pysat

import pysatSpaceWeather as pysat_sw


# --------------------------------------------------------------------------
# Instrument utilities

def acknowledgements(name, tag):
    """Returns acknowledgements for space weather dataset

    Parameters
    ----------
    name : string
        Name of space weather index, eg, dst, f107, kp
    tag : string
        Tag of the space waether index

    """
    swpc = ''.join(['Prepared by the U.S. Dept. of Commerce, NOAA, Space ',
                    'Weather Prediction Center'])

    ackn = {'kp': {'': 'Provided by GFZ German Research Centre for Geosciences',
                   'forecast': swpc, 'recent': swpc}}

    return ackn[name][tag]


def references(name, tag):
    """Returns references for space weather dataset

    Parameters
    ----------
    name : string
        Name of space weather index, eg, dst, f107, kp
    tag : string
        Tag of the space waether index

    """

    gen_refs = "\n".join([''.join(["J. Bartels, The technique of scaling ",
                                   "indices K and Q of geomagnetic activity, ",
                                   "Ann. Intern. Geophys. Year 4, 215-226, ",
                                   "1957."]),
                          ''.join(["J. Bartels,The geomagnetic measures for ",
                                   "the time-variations of solar corpuscular ",
                                   "radiation, described for use in ",
                                   "correlation studies in other geophysical ",
                                   "fields, Ann. Intern. Geophys. Year 4, ",
                                   "227-236, 1957."]),
                          ''.join(["P.N. Mayaud, Derivation, Meaning and Use ",
                                   "of Geomagnetic Indices, Geophysical ",
                                   "Monograph 22, Am. Geophys. Union, ",
                                   "Washington D.C., 1980."]),
                          ''.join(["G.K. Rangarajan, Indices of magnetic ",
                                   "activity, in Geomagnetism, edited by I.A. ",
                                   "Jacobs, Academic, San Diego, 1989."]),
                          ''.join(["M. Menvielle and A. Berthelier, The ",
                                   "K-derived planetary indices: description ",
                                   "and availability, Rev. Geophys. 29, 3, ",
                                   "415-432, 1991."])])
    refs = {'kp': {'': gen_refs, 'forecast': gen_refs, 'recent': gen_refs}}

    return refs[name][tag]


def initialize_kp_metadata(meta, data_key, fill_val=-1):
    """ Initialize the Kp meta data using our knowledge of the index

    Parameters
    ----------
    meta : pysat._meta.Meta
        Pysat Metadata
    data_key : str
        String denoting the data key
    fill_val : int or float
        File-specific fill value (default=-1)

    """

    meta[data_key] = {meta.labels.name: data_key,
                      meta.labels.desc: "Planetary K-index",
                      meta.labels.min_val: 0,
                      meta.labels.max_val: 9,
                      meta.labels.fill_val: fill_val}

    return


# --------------------------------------------------------------------------
# Common custom functions


def convert_ap_to_kp(ap_data, fill_val=-1, ap_name='ap'):
    """ Convert Ap into Kp

    Parameters
    ----------
    ap_data : array-like
        Array-like object containing Ap data
    fill_val : int, float, NoneType
        Fill value for the data set (default=-1)
    ap_name : str
        Name of the input ap

    Returns
    -------
    kp_data : array-like
        Array-like object containing Kp data
    meta : Metadata
        Metadata object containing information about transformed data

    """

    # Ap are keys, Kp returned as double (N- = N.6667, N+=N.3333333)
    one_third = 1.0 / 3.0
    two_third = 2.0 / 3.0
    ap_to_kp = {0: 0, 2: one_third, 3: two_third,
                4: 1, 5: 1.0 + one_third, 6: 1.0 + two_third,
                7: 2, 9: 2.0 + one_third, 12: 2.0 + two_third,
                15: 3, 18: 3.0 + one_third, 22: 3.0 + two_third,
                27: 4, 32: 4.0 + one_third, 39: 4.0 + two_third,
                48: 5, 56: 5.0 + one_third, 67: 5.0 + two_third,
                80: 6, 94: 6.0 + one_third, 111: 6.0 + two_third,
                132: 7, 154: 7.0 + one_third, 179: 7.0 + two_third,
                207: 8, 236: 8.0 + one_third, 300: 8.0 + two_third,
                400: 9}
    ap_keys = sorted([akey for akey in ap_to_kp.keys()])

    # If the ap falls between two Kp indices, assign it to the lower Kp value
    def round_ap(ap_in, fill_val=fill_val):
        """ Round an ap value to the nearest Kp value
        """
        if not np.isfinite(ap_in):
            return fill_val

        i = 0
        while ap_keys[i] <= ap_in:
            i += 1
        i -= 1

        if i >= len(ap_keys) or ap_keys[i] > ap_in:
            return fill_val

        return ap_to_kp[ap_keys[i]]

    # Convert from ap to Kp
    kp_data = np.array([ap_to_kp[aa] if aa in ap_keys else
                        round_ap(aa, fill_val=fill_val) for aa in ap_data])

    # Set the metadata
    meta = pysat.Meta()
    meta['Kp'] = {meta.labels.name: 'Kp',
                  meta.labels.desc: 'Kp converted from {:}'.format(ap_name),
                  meta.labels.min_val: 0,
                  meta.labels.max_val: 9,
                  meta.labels.fill_val: fill_val,
                  meta.labels.notes: ''.join(
                      ['Kp converted from ', ap_name, 'as described at: ',
                       'https://www.ngdc.noaa.gov/stp/GEOMAG/kp_ap.html'])}

    # Return data and metadata
    return(kp_data, meta)


def convert_3hr_kp_to_ap(kp_inst):
    """ Calculate 3 hour ap from 3 hour Kp index

    Parameters
    ----------
    kp_inst : pysat.Instrument
        Pysat instrument containing Kp data

    Notes
    -----
    Conversion between ap and Kp indices is described at:
    https://www.ngdc.noaa.gov/stp/GEOMAG/kp_ap.html

    """

    # Kp are keys, where n.3 = n+ and n.6 = (n+1)-. E.g., 0.6 = 1-
    kp_to_ap = {0: 0, 0.3: 2, 0.6: 3, 1: 4, 1.3: 5, 1.6: 6, 2: 7, 2.3: 9,
                2.6: 12, 3: 15, 3.3: 18, 3.6: 22, 4: 27, 4.3: 32, 4.6: 39,
                5: 48, 5.3: 56, 5.6: 67, 6: 80, 6.3: 94, 6.6: 111, 7: 132,
                7.3: 154, 7.6: 179, 8: 207, 8.3: 236, 8.6: 300, 9: 400}

    def ap(kk):
        if np.isfinite(kk):
            return kp_to_ap[np.floor(kk * 10.0) / 10.0]
        else:
            return np.nan

    # Test the input
    if 'Kp' not in kp_inst.data.columns:
        raise ValueError('unable to locate Kp data')

    # Convert from Kp to ap
    fill_val = kp_inst.meta['Kp'][kp_inst.meta.labels.fill_val]
    ap_data = np.array([ap(kp) if kp != fill_val else fill_val
                        for kp in kp_inst['Kp']])

    # Append the output to the pysat instrument
    kp_inst['3hr_ap'] = pds.Series(ap_data, index=kp_inst.index)

    # Add metadata
    meta_dict = {kp_inst.meta.labels.name: 'ap',
                 kp_inst.meta.labels.desc: "3-hour ap (equivalent range) index",
                 kp_inst.meta.labels.min_val: 0,
                 kp_inst.meta.labels.max_val: 400,
                 kp_inst.meta.labels.fill_val: fill_val,
                 kp_inst.meta.labels.notes: ''.join(
                     ['ap converted from Kp as described at: ',
                      'at: https://www.ngdc.noaa.gov/stp/GEOMAG/kp_ap.html'])}

    kp_inst.meta['3hr_ap'] = meta_dict


def calc_daily_Ap(ap_inst, ap_name='3hr_ap', daily_name='Ap',
                  running_name=None):
    """ Calculate the daily Ap index from the 3hr ap index

    Parameters
    ----------
    ap_inst : pysat.Instrument
        pysat instrument containing 3-hourly ap data
    ap_name : str
        Column name for 3-hourly ap data (default='3hr_ap')
    daily_name : str
        Column name for daily Ap data (default='Ap')
    running_name : str or NoneType
        Column name for daily running average of ap, not output if None
        (default=None)

    Note
    ----
    Ap is the mean of the 3hr ap indices measured for a given day

    Option for running average is included since this information is used
    by MSIS when running with sub-daily geophysical inputs

    """

    # Test that the necessary data is available
    if ap_name not in ap_inst.data.columns:
        raise ValueError("bad 3-hourly ap column name: {:}".format(ap_name))

    # Test to see that we will not be overwritting data
    if daily_name in ap_inst.data.columns:
        raise ValueError("daily Ap column name already exists: " + daily_name)

    # Calculate the daily mean value
    ap_mean = ap_inst[ap_name].rolling(window='1D', min_periods=8).mean()

    if running_name is not None:
        ap_inst[running_name] = ap_mean

        meta_dict = {ap_inst.meta.labels.name: running_name,
                     ap_inst.meta.labels.desc: "running daily Ap index",
                     ap_inst.meta.labels.min_val: 0,
                     ap_inst.meta.labels.max_val: 400,
                     ap_inst.meta.labels.fill_val:
                     ap_inst.meta[ap_name][ap_inst.meta.labels.fill_val],
                     ap_inst.meta.labels.notes:
                     '24-h running average of 3-hourly ap indices'}

        ap_inst.meta[running_name] = meta_dict

    # Resample, backfilling so that each day uses the mean for the data from
    # that day only
    #
    # Pad the data so the first day will be backfilled
    ap_pad = pds.Series(np.full(shape=(1,), fill_value=np.nan),
                        index=[ap_mean.index[0] - pds.DateOffset(hours=3)])
    # Extract the mean that only uses data for one day
    ap_sel = ap_pad.combine_first(ap_mean[[i for i, tt in
                                           enumerate(ap_mean.index)
                                           if tt.hour == 21]])
    # Backfill this data
    ap_data = ap_sel.resample('3H').backfill()

    # Save the output for the original time range
    ap_inst[daily_name] = pds.Series(ap_data[1:], index=ap_data.index[1:])

    # Add metadata
    meta_dict = {ap_inst.meta.labels.units: '',
                 ap_inst.meta.labels.name: 'Ap',
                 ap_inst.meta.labels.desc: "daily Ap index",
                 ap_inst.meta.labels.min_val: 0,
                 ap_inst.meta.labels.max_val: 400,
                 ap_inst.meta.labels.fill_val:
                 ap_inst.meta[ap_name][ap_inst.meta.labels.fill_val],
                 ap_inst.meta.labels.notes: ''.join(['Ap daily mean calculated',
                                                     ' from 3-hourly ap ',
                                                     'indices'])}

    ap_inst.meta[daily_name] = meta_dict

    return


def filter_geomag(inst, min_kp=0, max_kp=9, filter_time=24, kp_inst=None):
    """Filters pysat.Instrument data for given time after Kp drops below gate.

    Parameters
    ----------
    inst : pysat.Instrument
        Instrument with non-Kp data to be filtered by geomagnetic activity
    min_kp : float
        Minimum Kp value allowed. Kp values below this filter the data in
        inst (default=0)
    max_kp : float
        Maximum Kp value allowed. Kp values above this filter the data in
        inst (default=9)
    filter_time : int
        Number of hours to filter data after Kp drops below max_kp (default=24)
    kp_inst : pysat.Instrument or NoneType
        Kp pysat.Instrument object with or without data already loaded. If None,
        will load GFZ historic kp data for the instrument date (default=None)

    Note
    ----
    Loads Kp data for the same timeframe covered by inst and sets inst.data to
    NaN for times when Kp > max_kp or Kp < min_kp and for filter_time after Kp
    drops below max_kp.

    Default max and min values accept all Kp, so changing only one will cause
    the filter to act as a high- or low-pass function.

    This routine is written for standard Kp data (tag=''), not the forecast or
    recent data.  However, it will work with these Kp data if they are supplied.

    This routine is designed to be used with the 'modify' flag if applied as
    a custom routine.

    """
    # Load the desired data
    if kp_inst is None:
        kp_inst = pysat.Instrument(inst_module=pysat_sw.instruments.sw_kp,
                                   tag='', pad=pds.DateOffset(days=1))

    if kp_inst.empty:
        kp_inst.load(date=inst.date, verifyPad=True)

    # Begin filtering, starting at the beginning of the instrument data
    sel_data = kp_inst[(inst.date - pds.DateOffset(days=1)):
                       (inst.date + pds.DateOffset(days=1))]
    ind, = np.where((sel_data['Kp'] > max_kp) | (sel_data['Kp'] < min_kp))
    for lind in ind:
        sind = sel_data.index[lind]
        eind = sind + pds.DateOffset(hours=filter_time)
        inst[sind:eind] = np.nan
        inst.data = inst.data.dropna(axis=0, how='all')

    return


# --------------------------------------------------------------------------
# Common analysis functions


def combine_kp(standard_inst=None, recent_inst=None, forecast_inst=None,
               start=None, stop=None, fill_val=np.nan):
    """ Combine the output from the different Kp sources for a range of dates

    Parameters
    ----------
    standard_inst : pysat.Instrument or NoneType
        Instrument object containing data for the 'sw' platform, 'kp' name,
        and '' tag or None to exclude (default=None)
    recent_inst : pysat.Instrument or NoneType
        Instrument object containing data for the 'sw' platform, 'kp' name,
        and 'recent' tag or None to exclude (default=None)
    forecast_inst : pysat.Instrument or NoneType
        Instrument object containing data for the 'sw' platform, 'kp' name,
        and 'forecast' tag or None to exclude (default=None)
    start : dt.datetime or NoneType
        Starting time for combining data, or None to use earliest loaded
        date from the pysat Instruments (default=None)
    stop : dt.datetime
        Ending time for combining data, or None to use the latest loaded date
        from the pysat Instruments (default=None)
    fill_val : int or float
        Desired fill value (since the standard instrument fill value differs
        from the other sources) (default=np.nan)

    Returns
    -------
    kp_inst : pysat.Instrument
        Instrument object containing Kp observations for the desired period of
        time, merging the standard, recent, and forecasted values based on
        their reliability

    Note
    ----
    Merging prioritizes the standard data, then the recent data, and finally
    the forecast data

    Will not attempt to download any missing data, but will load data

    """

    notes = "Combines data from"

    # Create an ordered list of the Instruments, excluding any that are None
    all_inst = list()
    tag = 'combined'
    inst_flag = None
    if standard_inst is not None:
        all_inst.append(standard_inst)
        tag += '_standard'
        if inst_flag is None:
            inst_flag = 'standard'

    if recent_inst is not None:
        all_inst.append(recent_inst)
        tag += '_recent'
        if inst_flag is None:
            inst_flag = 'recent'

    if forecast_inst is not None:
        all_inst.append(forecast_inst)
        tag += '_forecast'
        if inst_flag is None:
            inst_flag = 'forecast'

    if len(all_inst) < 2:
        raise ValueError("need at two Kp Instrument objects to combine them")

    # If the start or stop times are not defined, get them from the Instruments
    if start is None:
        stimes = [inst.index.min() for inst in all_inst if len(inst.index) > 0]
        start = min(stimes) if len(stimes) > 0 else None

    if stop is None:
        stimes = [inst.index.max() for inst in all_inst if len(inst.index) > 0]
        stop = max(stimes) if len(stimes) > 0 else None
        stop += pds.DateOffset(days=1)

    if start is None or stop is None:
        raise ValueError(' '.join(("must either load in Instrument objects or",
                                   "provide starting and ending times")))

    # Initialize the output instrument
    kp_inst = pysat.Instrument(labels=all_inst[0].meta_labels)
    kp_inst.inst_module = pysat_sw.instruments.sw_kp
    kp_inst.tag = tag
    kp_inst.date = start
    kp_inst.doy = int(start.strftime("%j"))
    kp_inst.meta = pysat.Meta(labels=kp_inst.meta_labels)
    initialize_kp_metadata(kp_inst.meta, 'Kp', fill_val=fill_val)

    kp_times = list()
    kp_values = list()

    # Cycle through the desired time range
    itime = start
    while itime < stop and inst_flag is not None:
        # Load and save the standard data for as many times as possible
        if inst_flag == 'standard':
            standard_inst.load(date=itime)

            if notes.find("standard") < 0:
                notes += " the {:} source ({:} to ".format(inst_flag,
                                                           itime.date())

            if len(standard_inst.index) == 0:
                inst_flag = 'forecast' if recent_inst is None else 'recent'
                notes += "{:})".format(itime.date())
            else:
                kp_times.extend(list(standard_inst.index))
                kp_values.extend(list(standard_inst['Kp']))
                itime = kp_times[-1] + pds.DateOffset(hours=3)

        # Load and save the recent data for as many times as possible
        if inst_flag == 'recent':
            # Determine which files should be loaded
            if len(recent_inst.index) == 0:
                files = np.unique(recent_inst.files.files[itime:stop])
            else:
                files = [None]  # No load needed, if already initialized

            # Cycle through all possible files of interest, saving relevant
            # data
            for filename in files:
                if filename is not None:
                    recent_inst.load(fname=filename)

                if notes.find("recent") < 0:
                    notes += " the {:} source ({:} to ".format(inst_flag,
                                                               itime.date())

                # Determine which times to save
                local_fill_val = recent_inst.meta[
                    'Kp', recent_inst.meta.labels.fill_val]
                good_times = ((recent_inst.index >= itime)
                              & (recent_inst.index < stop))
                good_vals = recent_inst['Kp'][good_times] != local_fill_val

                # Save output data and cycle time
                kp_times.extend(list(recent_inst.index[good_times][good_vals]))
                kp_values.extend(list(recent_inst['Kp'][good_times][good_vals]))
                itime = kp_times[-1] + pds.DateOffset(hours=3)

            inst_flag = 'forecast' if forecast_inst is not None else None
            notes += "{:})".format(itime.date())

        # Load and save the forecast data for as many times as possible
        if inst_flag == "forecast":
            # Determine which files should be loaded
            if len(forecast_inst.index) == 0:
                files = np.unique(forecast_inst.files.files[itime:stop])
            else:
                files = [None]  # No load needed, if already initialized

            # Cycle through all possible files of interest, saving relevant
            # data
            for filename in files:
                if filename is not None:
                    forecast_inst.load(fname=filename)

                if notes.find("forecast") < 0:
                    notes += " the {:} source ({:} to ".format(inst_flag,
                                                               itime.date())

                # Determine which times to save
                local_fill_val = forecast_inst.meta[
                    'Kp', forecast_inst.meta.labels.fill_val]
                good_times = ((forecast_inst.index >= itime)
                              & (forecast_inst.index < stop))
                good_vals = forecast_inst['Kp'][good_times] != local_fill_val

                # Save desired data and cycle time
                new_times = list(forecast_inst.index[good_times][good_vals])
                kp_times.extend(new_times)
                new_vals = list(forecast_inst['Kp'][good_times][good_vals])
                kp_values.extend(new_vals)
                itime = kp_times[-1] + pds.DateOffset(hours=3)
            notes += "{:})".format(itime.date())

            inst_flag = None

    if inst_flag is not None:
        notes += "{:})".format(itime.date())

    # Determine if the beginning or end of the time series needs to be padded

    freq = None if len(kp_times) < 2 else pysat.utils.time.calc_freq(kp_times)
    end_date = stop - pds.DateOffset(days=1)
    date_range = pds.date_range(start=start, end=end_date, freq=freq)

    if len(kp_times) == 0:
        kp_times = date_range

    if date_range[0] < kp_times[0]:
        # Extend the time and value arrays from their beginning with fill
        # values
        itime = abs(date_range - kp_times[0]).argmin()
        kp_times.reverse()
        kp_values.reverse()
        extend_times = list(date_range[:itime])
        extend_times.reverse()
        kp_times.extend(extend_times)
        kp_values.extend([fill_val for kk in extend_times])
        kp_times.reverse()
        kp_values.reverse()

    if date_range[-1] > kp_times[-1]:
        # Extend the time and value arrays from their end with fill values
        itime = abs(date_range - kp_times[-1]).argmin() + 1
        extend_times = list(date_range[itime:])
        kp_times.extend(extend_times)
        kp_values.extend([fill_val for kk in extend_times])

    # Save output data
    kp_inst.data = pds.DataFrame(kp_values, columns=['Kp'], index=kp_times)

    # Resample the output data, filling missing values
    if (date_range.shape != kp_inst.index.shape
            or abs(date_range - kp_inst.index).max().total_seconds() > 0.0):
        kp_inst.data = kp_inst.data.resample(freq).fillna(method=None)
        if np.isfinite(fill_val):
            kp_inst.data[np.isnan(kp_inst.data)] = fill_val

    # Update the metadata notes for this custom procedure
    notes += ", in that order"
    kp_inst.meta['Kp', kp_inst.meta.labels.notes] = notes

    return kp_inst



# -*- coding: utf-8 -*-
"""Supports Kp index values. Downloads data from ftp.gfz-potsdam.de or SWPC.

Properties
----------
platform
    'sw'
name
    'kp'
tag
    - '' Standard Kp data
    - 'forecast' Grab forecast data from SWPC (next 3 days)
    - 'recent' Grab last 30 days of Kp data from SWPC

Note
----
Standard Kp files are stored by the first day of each month. When downloading
use kp.download(start, stop, freq='MS') to only download days that could
possibly have data.  'MS' gives a monthly start frequency.

The forecast data is stored by generation date, where each file contains the
forecast for the next three days. Forecast data downloads are only supported
for the current day. When loading forecast data, the date specified with the
load command is the date the forecast was generated. The data loaded will span
three days. To always ensure you are loading the most recent data, load
the data with tomorrow's date.
::

    kp = pysat.Instrument('sw', 'kp', tag='recent')
    kp.download()
    kp.load(date=kp.tomorrow())

Recent data is also stored by the generation date from the SWPC. Each file
contains 30 days of Kp measurements. The load date issued to pysat corresponds
to the generation date.

The recent and forecast data should not be used with the data padding option
available from pysat.Instrument objects.

Warnings
--------
The 'forecast' Kp data loads three days at a time. The data padding feature
and multi_file_day feature available from the pyast.Instrument object
is not appropriate for Kp 'forecast' data.


This material is based upon work supported by the
National Science Foundation under Grant Number 1259508.

Any opinions, findings, and conclusions or recommendations expressed in this
material are those of the author(s) and do not necessarily reflect the views
of the National Science Foundation.

"""

import datetime as dt
import ftplib
import numpy as np
import os
import pandas as pds
import requests
import sys

import pysat
from pysat.utils.time import parse_date

from pysatSpaceWeather.instruments.methods import kp_ap

logger = pysat.logger

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'kp'
tags = {'': '',
        'forecast': 'SWPC Forecast data next (3 days)',
        'recent': 'SWPC provided Kp for past 30 days'}
inst_ids = {'': ['', 'forecast', 'recent']}

# generate todays date to support loading forecast data
now = dt.datetime.utcnow()
today = dt.datetime(now.year, now.month, now.day)

# ----------------------------------------------------------------------------
# Instrument test attributes

# Set test dates
_test_dates = {'': {'': dt.datetime(2009, 1, 1),
                    'forecast': today + dt.timedelta(days=1),
                    'recent': today}}

# Other tags assumed to be True
_test_download_travis = {'': {'': False}}

# ----------------------------------------------------------------------------
# Instrument methods


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    """

    self.acknowledgements = kp_ap.acknowledgements(self.name, self.tag)
    self.references = kp_ap.references(self.name, self.tag)
    logger.info(self.acknowledgements)
    return


def clean(self):
    """ Cleaning function for Space Weather indices

    Note
    ----
    Kp doesn't require cleaning
    """
    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag=None, inst_id=None):
    """Load Kp index files

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames
    tag : str or NoneType
        tag or None (default=None)
    inst_id : str or NoneType
        satellite id or None (default=None)

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data
    meta : pysat.Meta
        Object containing metadata such as column names and units

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    meta = pysat.Meta()
    if tag == '':
        # Kp data stored monthly, need to return data daily.  The daily date is
        # attached to filename.  Parse off the last date, load month of data,
        # and downselect to the desired day
        data = pds.DataFrame()

        # Set up fixed width format for these files
        colspec = [(0, 2), (2, 4), (4, 6), (7, 10), (10, 13), (13, 16),
                   (16, 19), (19, 23), (23, 26), (26, 29), (29, 32), (32, 50)]
        all_data = []
        for filename in fnames:
            # The daily date is attached to the filename.  Parse off the last
            # date, load month of data, downselect to the desired day
            fname = filename[0:-11]
            date = dt.datetime.strptime(filename[-10:], '%Y-%m-%d')

            temp = pds.read_fwf(fname, colspecs=colspec, skipfooter=4,
                                header=None, parse_dates=[[0, 1, 2]],
                                date_parser=parse_date, index_col='0_1_2')
            idx, = np.where((temp.index >= date)
                            & (temp.index < date + pds.DateOffset(days=1)))
            temp = temp.iloc[idx, :]
            all_data.append(temp)

        # Combine data together
        data = pds.concat(all_data, axis=0, sort=True)

        # Drop last column as it has data that belongs to a different inst
        data = data.iloc[:, 0:-1]

        # Each column increments UT by three hours. Produce a single data
        # series that has Kp value monotonically increasing in time with
        # appropriate datetime indices
        data_series = pds.Series(dtype='float64')
        for i in np.arange(8):
            tind = data.index + pds.DateOffset(hours=int(3 * i))
            temp = pds.Series(data.iloc[:, i].values, index=tind)
            data_series = data_series.append(temp)
        data_series = data_series.sort_index()
        data_series.index.name = 'time'

        # Kp comes in non-user friendly values like 2-, 2o, and 2+. Relate
        # these to 1.667, 2.0, 2.333 for processing and user friendliness
        first = np.array([float(x[0]) for x in data_series])
        flag = np.array([x[1] for x in data_series])

        ind, = np.where(flag == '+')
        first[ind] += 1.0 / 3.0
        ind, = np.where(flag == '-')
        first[ind] -= 1.0 / 3.0

        result = pds.DataFrame(first, columns=['Kp'], index=data_series.index)
        fill_val = np.nan
    elif tag == 'forecast':
        # Load forecast data
        all_data = []
        for fname in fnames:
            result = pds.read_csv(fname, index_col=0, parse_dates=True)
            all_data.append(result)
        result = pds.concat(all_data)
        fill_val = -1
    elif tag == 'recent':
        # Load recent Kp data
        all_data = []
        for fname in fnames:
            result = pds.read_csv(fname, index_col=0, parse_dates=True)
            all_data.append(result)
        result = pds.concat(all_data)
        fill_val = -1

    # Initalize the meta data
    for kk in result.keys():
        kp_ap.initialize_kp_metadata(meta, kk, fill_val)

    return result, meta


def list_files(tag=None, inst_id=None, data_path=None, format_str=None):
    """Return a Pandas Series of every file for chosen satellite data

    Parameters
    -----------
    tag : string or NoneType
        Denotes type of file to load.
        (default=None)
    inst_id : string or NoneType
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    data_path : string or NoneType
        Path to data directory.  If None is specified, the value previously
        set in Instrument.files.data_path is used.  (default=None)
    format_str : string or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    pysat.Files.from_os : pysat._files.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    if data_path is not None:
        if tag == '':
            # Files are by month, going to add date to monthly filename for
            # each day of the month. The load routine will load a month of
            # data and use the appended date to select out appropriate data.
            if format_str is None:
                format_str = 'kp{year:2d}{month:02d}.tab'
            out = pysat.Files.from_os(data_path=data_path,
                                      format_str=format_str,
                                      two_digit_year_break=99)
            if not out.empty:
                out.loc[out.index[-1] + pds.DateOffset(months=1)
                        - pds.DateOffset(days=1)] = out.iloc[-1]
                out = out.asfreq('D', 'pad')
                out = out + '_' + out.index.strftime('%Y-%m-%d')
            return out
        elif tag == 'forecast':
            format_str = 'kp_forecast_{year:04d}-{month:02d}-{day:02d}.txt'
            files = pysat.Files.from_os(data_path=data_path,
                                        format_str=format_str)
            # Pad list of files data to include most recent file under tomorrow
            if not files.empty:
                pds_offset = pds.DateOffset(days=1)
                files.loc[files.index[-1] + pds_offset] = files.values[-1]
                files.loc[files.index[-1] + pds_offset] = files.values[-1]
            return files
        elif tag == 'recent':
            format_str = 'kp_recent_{year:04d}-{month:02d}-{day:02d}.txt'
            files = pysat.Files.from_os(data_path=data_path,
                                        format_str=format_str)

            # Pad list of files data to include most recent file under tomorrow
            if not files.empty:
                pds_offset = pds.DateOffset(days=1)
                files.loc[files.index[-1] + pds_offset] = files.values[-1]
                files.loc[files.index[-1] + pds_offset] = files.values[-1]
            return files

        else:
            raise ValueError(' '.join(('Unrecognized tag name for Space',
                                       'Weather Index Kp')))
    else:
        raise ValueError(' '.join(('A data_path must be passed to the loading',
                                   'routine for Kp')))


def download(date_array, tag, inst_id, data_path, user=None, password=None):
    """Routine to download Kp index data

    Parameters
    -----------
    tag : string or NoneType
        Denotes type of file to load.  Accepted types are '' and 'forecast'.
        (default=None)
    inst_id : string or NoneType
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    data_path : string or NoneType
        Path to data directory.  If None is specified, the value previously
        set in Instrument.files.data_path is used.  (default=None)

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    Only able to download current forecast data, not archived forecasts.

    """

    # Download standard Kp data
    if tag == '':
        ftp = ftplib.FTP('ftp.gfz-potsdam.de')  # connect to host, default port
        ftp.login()  # user anonymous, passwd anonymous@
        ftp.cwd('/pub/home/obs/kp-ap/tab')
        dnames = list()

        for date in date_array:
            fname = 'kp{year:02d}{month:02d}.tab'
            fname = fname.format(year=(date.year - (date.year // 100) * 100),
                                 month=date.month)
            local_fname = fname
            saved_fname = os.path.join(data_path, local_fname)
            if fname not in dnames:
                try:
                    logger.info(' '.join(('Downloading file for',
                                          date.strftime('%b %Y'))))
                    sys.stdout.flush()
                    ftp.retrbinary('RETR ' + fname,
                                   open(saved_fname, 'wb').write)
                    dnames.append(fname)
                except ftplib.error_perm as exception:

                    if str(exception.args[0]).split(" ", 1)[0] != '550':
                        # Leaving a bare raise below so that ftp errors
                        # are properly reported as coming from ftp
                        # and gives the correct line number.
                        # We aren't expecting any 'normal' ftp errors
                        # here, other than a 550 'no file' error, thus
                        # accurately raising FTP issues is the way to go
                        raise
                    else:
                        # File isn't actually there, just let people know
                        # then continue on
                        os.remove(saved_fname)
                        logger.info(' '.join(('File not available for',
                                              date.strftime('%x'))))

        ftp.close()

    elif tag == 'forecast':
        logger.info(' '.join(('This routine can only download the current',
                              'forecast, not archived forecasts')))
        # Download webpage
        furl = 'https://services.swpc.noaa.gov/text/3-day-geomag-forecast.txt'
        r = requests.get(furl)

        # Parse text to get the date the prediction was generated
        date_str = r.text.split(':Issued: ')[-1].split(' UTC')[0]
        date = dt.datetime.strptime(date_str, '%Y %b %d %H%M')

        # Data is the forecast value for the next three days
        raw_data = r.text.split('NOAA Kp index forecast ')[-1]

        # Get date of the forecasts
        date_str = raw_data[0:6] + ' ' + str(date.year)
        forecast_date = dt.datetime.strptime(date_str, '%d %b %Y')

        # Strings we will use to parse the downloaded text
        lines = ['00-03UT', '03-06UT', '06-09UT', '09-12UT', '12-15UT',
                 '15-18UT', '18-21UT', '21-00UT']

        # Storage for daily forecasts.
        # Get values for each day, then combine together
        day1 = []
        day2 = []
        day3 = []
        for line in lines:
            raw = raw_data.split(line)[-1].split('\n')[0]
            day1.append(int(raw[0:10]))
            day2.append(int(raw[10:20]))
            day3.append(int(raw[20:]))
        times = pds.date_range(forecast_date, periods=24, freq='3H')
        day = []
        for dd in [day1, day2, day3]:
            day.extend(dd)

        # Put data into nicer DataFrame
        data = pds.DataFrame(day, index=times, columns=['Kp'])

        # Write out as a file
        data_file = 'kp_forecast_{:s}.txt'.format(date.strftime('%Y-%m-%d'))
        data.to_csv(os.path.join(data_path, data_file), header=True)

    elif tag == 'recent':
        logger.info(' '.join(('This routine can only download the current',
                              'webpage, not archived forecasts')))

        # Download webpage
        rurl = ''.join(('https://services.swpc.noaa.gov/text/',
                        'daily-geomagnetic-indices.txt'))
        r = requests.get(rurl)

        # Parse text to get the date the prediction was generated
        date_str = r.text.split(':Issued: ')[-1].split('\n')[0]
        date = dt.datetime.strptime(date_str, '%H%M UT %d %b %Y')

        # Data is the forecast value for the next three days
        raw_data = r.text.split('#  Date ')[-1]

        # Keep only the middle bits that matter
        raw_data = raw_data.split('\n')[1:-1]

        # Hold times from the file
        kp_time = []

        # Holds Kp value for each station
        sub_kps = [[], [], []]

        # Iterate through file lines and parse out the info we want
        for line in raw_data:
            kp_time.append(dt.datetime.strptime(line[0:10], '%Y %m %d'))

            # Pick out Kp values for each of the three columns
            sub_lines = [line[17:33], line[40:56], line[63:]]
            for sub_line, sub_kp in zip(sub_lines, sub_kps):
                for i in np.arange(8):
                    sub_kp.append(int(sub_line[(i * 2):((i + 1) * 2)]))

        # Create times on 3 hour cadence
        times = pds.date_range(kp_time[0], periods=(8 * 30), freq='3H')

        # Put into DataFrame
        data = pds.DataFrame({'mid_lat_Kp': sub_kps[0],
                              'high_lat_Kp': sub_kps[1],
                              'Kp': sub_kps[2]}, index=times)

        # Write out as a file
        data_file = 'kp_recent_{:s}.txt'.format(date.strftime('%Y-%m-%d'))
        data.to_csv(os.path.join(data_path, data_file), header=True)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2020, Authors
# Full license can be found in License.md
# -----------------------------------------------------------------------------

from setuptools import setup


# Run setup.  Setuptools will look for parameters in [metadata] section of
# setup.cfg
setup()



# -*- coding: utf-8 -*-.
"""Provides default routines for Dst

"""


def acknowledgements(name, tag):
    """Returns acknowledgements for space weather dataset

    Parameters
    ----------
    name : string
        Name of space weather index, eg, dst, f107, kp
    tag : string
        Tag of the space weather index

    """

    ackn = {'dst': {'': 'Dst is maintained at NCEI (formerly NGDC) at NOAA'}}

    return ackn[name][tag]


def references(name, tag):
    """Returns references for space weather dataset

    Parameters
    ----------
    name : string
        Name of space weather index, eg, dst, f107, kp
    tag : string
        Tag of the space weather index

    """

    refs = {'dst': {'': ''.join([
        'See referenece list and publication at: Sugiura M. and T. Kamei, '
        'http://wdc.kugi.kyoto-u.ac.jp/dstdir/dst2/onDstindex.html, ',
        'last updated June 1991, accessed Dec 2020'])}}

    return refs[name][tag]
